from __future__ import annotations

from typing import Iterable

import numpy as np
import pandas as pd

from psyke import get_default_random_seed, Target
from psyke.extraction.hypercubic import Grid, HyperCube, GenericCube, ClassificationCube
from psyke.extraction.hypercubic.gridex import GridEx


class HEx(GridEx):
    """
    Explanator implementing HEx algorithm.
    """

    class Node:
        def __init__(self, cube: GenericCube, parent: HEx.Node = None, threshold: float = None):
            self.cube = cube
            self.parent = parent
            self.children: Iterable[HEx.Node] = []
            self.threshold = threshold
            self.gain = True if parent is None else self.check()

        def check(self) -> bool:
            other = self.parent
            try:
                while not other.gain:
                    other = other.parent
            except AttributeError:
                return True
            if isinstance(other.cube, ClassificationCube):
                return other.cube.output != self.cube.output
            return other.cube.error - self.cube.error > self.threshold * .6

        def indices(self, dataframe: pd.DataFrame):
            return self.cube.filter_indices(dataframe.iloc[:, :-1])

        def eligible_children(self, dataframe) -> Iterable[HEx.Node]:
            return [c for c in self.children if c.cube.count(dataframe) > 0]

        def permanent_children(self, dataframe) -> Iterable[HEx.Node]:
            return [c for c in self.eligible_children(dataframe) if c.gain]

        def permanent_indices(self, dataframe):
            return np.any([c.cube.filter_indices(dataframe.iloc[:, :-1])
                           for c in self.eligible_children(dataframe) if c.gain], axis=0)

        def update(self, dataframe: pd.DataFrame, predictor, recursive=False):
            if recursive:
                for node in self.children:
                    node.update(dataframe, predictor, recursive)
            cleaned = [(c.cube, c.gain) for c in self.eligible_children(dataframe)]
            idx = self.permanent_indices(dataframe)

            if sum(g for _, g in cleaned) > 0 and sum(self.indices(dataframe)) > sum(idx) and self.gain:
                self.cube.update(dataframe[self.indices(dataframe) & ~idx], predictor)
            return cleaned

        def linearize(self, dataframe, depth=1):
            children = [c.linearize(dataframe, depth + 1) for c in self.permanent_children(dataframe)]
            return [(cc, dd) for c in children for cc, dd in c if c != []] + \
                   [(c, depth) for c in self.permanent_children(dataframe)]

    def __init__(self, predictor, grid: Grid, min_examples: int, threshold: float, output: Target = Target.CONSTANT,
                 discretization=None, normalization=None, seed: int = get_default_random_seed()):
        super().__init__(predictor, grid, min_examples, threshold, output, discretization, normalization, seed)
        self._default_surrounding_cube = True

    def _gain(self, parent_cube: GenericCube, new_cube: GenericCube) -> float:
        if isinstance(parent_cube, ClassificationCube):
            return parent_cube.output != new_cube.output
        return parent_cube.error - new_cube.error > self.threshold * .6

    def _iterate(self, dataframe: pd.DataFrame):
        fake = dataframe.copy()
        self._surrounding.update(dataframe, self.predictor)
        root = HEx.Node(self._surrounding, threshold=self.threshold)
        current = [root]

        for iteration in self.grid.iterate():
            next_iteration = []
            for node in current:
                if node.cube.diversity < self.threshold:
                    continue
                children, fake = self._cubes_to_split(node.cube, iteration, dataframe, fake, True)
                node.children = [HEx.Node(c, node, threshold=self.threshold) for c in children]
                cleaned = node.update(fake, self.predictor, False)
                node.children = [HEx.Node(c, node, threshold=self.threshold) for c in self._merge(
                    [c for c, _ in cleaned], fake)]
                next_iteration += [n for n in node.children]

            current = next_iteration.copy()
        _ = root.update(fake, self.predictor, True)
        self._hypercubes = []
        linearized = root.linearize(fake)
        for depth in sorted(np.unique([d for (_, d) in linearized]), reverse=True):
            self._hypercubes += self._merge([c.cube for (c, d) in linearized if d == depth], fake)

        if len(self._hypercubes) == 0:
            self._hypercubes = [self._surrounding]
        elif not min(np.any([c.filter_indices(dataframe.iloc[:, :-1]) for c in self._hypercubes], axis=0)):
            self._hypercubes = self._hypercubes + [self._surrounding]



import numpy as np
import pandas as pd
from tuprolog.theory import Theory

from psyke import Target, get_default_random_seed
from psyke.extraction.hypercubic import HyperCubeExtractor
from psyke.extraction.hypercubic.hypercube import Point, GenericCube, HyperCube

from sklearn.neighbors import BallTree


class DiViNE(HyperCubeExtractor):
    """
    Explanator implementing DiViNE algorithm.
    """

    def __init__(self, predictor, k: int = 5, patience: int = 15, close_to_center: bool = True,
                 discretization=None, normalization=None, seed: int = get_default_random_seed()):
        super().__init__(predictor, Target.CLASSIFICATION, discretization, normalization)
        self.k = k
        self.patience = patience
        self.vicinity_function = DiViNE.closest_to_center if close_to_center else DiViNE.closest_to_corners
        self.seed = seed

    @staticmethod
    def __pop(data: pd.DataFrame, idx: int = None) -> (Point, pd.DataFrame):
        if idx is None:
            idx = data.sample(1).index.values[0]
        t = data.T
        return DiViNE.__to_point(t.pop(idx)), t.T.reset_index(drop=True)

    @staticmethod
    def __to_point(instance) -> Point:
        point = Point(instance.index.values, instance.values)
        return point

    def __to_cube(self, point: Point) -> GenericCube:
        cube = HyperCube.cube_from_point(point.dimensions, self._output)
        cube._output = list(point.dimensions.values())[-1]
        return cube

    def __clean(self, data: pd.DataFrame) -> pd.DataFrame:
        _, idx = BallTree(data.iloc[:, :-1]).query(data.iloc[:, :-1], k=self.k)
        # how many output classes are associated with the k neighbors
        count = np.array(list(map(lambda indices: len(data.iloc[indices].iloc[:, -1].unique()), idx)))
        # instances with neighbors of different classes are discarded
        return data[count == 1]

    def __closest(self, data: pd.DataFrame, cube: GenericCube) -> (Point, pd.DataFrame):
        return DiViNE.__pop(data, self.vicinity_function(BallTree(data.iloc[:, :-1]), cube))

    @staticmethod
    def closest_to_center(tree: BallTree, cube: GenericCube):
        return tree.query([list(cube.center.dimensions.values())], k=1)[1][0][-1]

    @staticmethod
    def closest_to_corners(tree: BallTree, cube: GenericCube):
        distance, idx = tree.query([list(point.dimensions.values()) for point in cube.corners()], k=1)
        return idx[np.argmin(distance)][-1]

    def _extract(self, dataframe: pd.DataFrame) -> Theory:
        self._surrounding = HyperCube.create_surrounding_cube(dataframe, output=Target.CLASSIFICATION)
        np.random.seed(self.seed)
        data = self.__clean(dataframe)

        while len(data) > 0:
            discarded = []
            patience = self.patience
            point, data = self.__pop(data)
            cube = self.__to_cube(point)

            while patience > 0 and len(data) > 0:
                other, data = self.__closest(data, cube)
                if cube.output == list(other.dimensions.values())[-1]:
                    cube = cube.merge_with_point(other)
                    data = data[~(cube.filter_indices(data.iloc[:, :-1]))].reset_index(drop=True)
                else:
                    patience -= 1
                    discarded.append(other)
            if cube.volume() > 0:
                cube.update(dataframe, self.predictor)
                self._hypercubes.append(cube)
            if len(discarded) > 0:
                data = pd.concat([data] + [d.to_dataframe() for d in discarded]).reset_index(drop=True)
        self._sort_cubes()
        return self._create_theory(dataframe)



from __future__ import annotations

from abc import ABC
from collections import Counter
from typing import Iterable, Union

import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor

from psyke.clustering import HyperCubeClustering
from psyke.extraction.hypercubic import Node, ClosedCube, HyperCube
from psyke.clustering.utils import select_gaussian_mixture, select_dbscan_epsilon
from psyke.extraction.hypercubic.hypercube import ClosedRegressionCube, ClosedClassificationCube
from psyke.utils import Target, get_default_random_seed


class ExACT(HyperCubeClustering, ABC):
    """
    Explanator implementing ExACT algorithm.
    """

    def __init__(self, depth: int = 2, error_threshold: float = 0.1, output: Target = Target.CONSTANT,
                 gauss_components: int = 2, discretization=None, normalization=None,
                 seed: int = get_default_random_seed()):
        super().__init__(output, discretization, normalization)
        self.depth = depth
        self.error_threshold = error_threshold
        self.gauss_components = gauss_components
        self._predictor = KNeighborsClassifier() if output == Target.CLASSIFICATION else KNeighborsRegressor()
        self._predictor.n_neighbors = 1
        self.seed = seed

    def __eligible_cubes(self, gauss_pred: np.ndarray, node: Node, clusters: int):
        cubes = []
        for i in range(len(np.unique(gauss_pred))):
            df = node.dataframe.iloc[np.where(gauss_pred == i)]
            if len(df) == 0:
                continue
            cubes.append(self._create_cube(df, clusters))
        indices = [self._indices(cube, node.dataframe) for cube in cubes]
        return cubes, indices

    @staticmethod
    def _indices(cube: ClosedCube, data: pd.DataFrame) -> np.ndarray | None:
        indices = cube.filter_indices(data.iloc[:, :-1])
        if len(data.iloc[indices]) * len(data.iloc[~indices]) == 0:
            return None
        return indices

    def _create_cube(self, dataframe: pd.DataFrame, clusters: int) -> ClosedCube:
        data = ExACT._remove_string_label(dataframe)
        dbscan_pred = DBSCAN(eps=select_dbscan_epsilon(data, clusters)).fit_predict(data.iloc[:, :-1])
        return HyperCube.create_surrounding_cube(
            dataframe.iloc[np.where(dbscan_pred == Counter(dbscan_pred).most_common(1)[0][0])],
            True, self._output
        )

    def fit(self, dataframe: pd.DataFrame):
        np.random.seed(self.seed)
        self._predictor.fit(dataframe.iloc[:, :-1], dataframe.iloc[:, -1])
        self._surrounding = HyperCube.create_surrounding_cube(dataframe, True, self._output)
        self._hypercubes = self._iterate(Node(dataframe, self._surrounding))

    def get_hypercubes(self) -> Iterable[HyperCube]:
        return list(self._hypercubes)

    def explain(self):
        for cube in self._hypercubes:
            print(f'Output is {cube.output} if:')
            for feature in cube.dimensions:
                lower, upper = cube[feature]
                print(f'    {feature} is in [{lower:.2f}, {upper:.2f}]')

    @staticmethod
    def _remove_string_label(dataframe: pd.DataFrame):
        return dataframe.replace({dataframe.columns[-1]: {v: k for k, v in dict(
            enumerate(dataframe.iloc[:, -1].unique())
        ).items()}}) if isinstance(dataframe.iloc[0, -1], str) else dataframe

    def _iterate(self, surrounding: Node) -> Iterable[HyperCube]:
        to_split = [(self.error_threshold * 10, 1, 1, surrounding)]
        while len(to_split) > 0:
            to_split.sort(reverse=True)
            (_, depth, _, node) = to_split.pop()
            data = ExACT._remove_string_label(node.dataframe)
            gauss_params = select_gaussian_mixture(data, self.gauss_components)
            gauss_pred = gauss_params[2].predict(data)
            cubes, indices = self.__eligible_cubes(gauss_pred, node, gauss_params[1])
            cubes = [(c.volume(), len(idx), i, idx, c) for i, (c, idx) in enumerate(zip(cubes, indices))
                     if (idx is not None) and (not node.cube.equal(c))]
            if len(cubes) < 1:
                continue
            _, _, _, indices, cube = max(cubes)

            cube.update(node.dataframe[indices], self._predictor)
            node.right = Node(node.dataframe[indices], cube)
            node.cube.update(node.dataframe[~indices], self._predictor)
            node.left = Node(node.dataframe[~indices], node.cube)

            if depth < self.depth and cube.diversity > self.error_threshold:
                to_split.append((cube.diversity, depth + 1, np.random.uniform(), node.right))
        return self._node_to_cubes(surrounding)

    def _node_to_cubes(self, root: Node) -> list[ClosedCube]:
        if root.right is None:
            return [root.cube]
        else:
            return self._node_to_cubes(root.right) + self._node_to_cubes(root.left)

    def _default_cube(self) -> Union[ClosedCube, ClosedRegressionCube, ClosedClassificationCube]:
        if self._output == Target.CONSTANT:
            return ClosedCube()
        if self._output == Target.REGRESSION:
            return ClosedRegressionCube()
        return ClosedClassificationCube()



import numpy as np
import pandas as pd
from sklearn.mixture import GaussianMixture
from tuprolog.theory import Theory

from psyke import Target, Extractor, get_default_random_seed
from psyke.clustering.utils import select_gaussian_mixture
from psyke.extraction.hypercubic import HyperCube, HyperCubeExtractor, RegressionCube


class COSMiK(HyperCubeExtractor):
    """
    Explanator implementing COSMiK algorithm.
    """

    def __init__(self, predictor, max_components: int = 4, k: int = 5, patience: int = 15, close_to_center: bool = True,
                 output: Target = Target.CONSTANT, discretization=None, normalization=None,
                 seed: int = get_default_random_seed()):
        super().__init__(predictor, Target.REGRESSION, discretization, normalization)
        self.max = max_components
        self.k = k
        self.patience = patience
        self.output = output
        self.close_to_center = close_to_center
        self.seed = seed

    def _extract(self, dataframe: pd.DataFrame) -> Theory:
        np.random.seed(self.seed)
        X, y = dataframe.iloc[:, :-1], dataframe.iloc[:, -1]

        _, n, _ = select_gaussian_mixture(dataframe, self.max)
        gmm = GaussianMixture(n)
        gmm.fit(X, y)

        divine = Extractor.divine(gmm, self.k, self.patience, self.close_to_center,
                                  self.discretization, self.normalization)
        df = X.join(pd.DataFrame(gmm.predict(X)))
        df.columns = dataframe.columns
        divine.extract(df)

        self._hypercubes = [HyperCube(cube.dimensions.copy()) if self.output == Target.CONSTANT else
                            RegressionCube(cube.dimensions.copy()) for cube in divine._hypercubes]
        for cube in self._hypercubes:
            cube.update(dataframe, self.predictor)

        self._sort_cubes()
        return self._create_theory(dataframe)


from array import array
from typing import Callable, Iterable
import numpy as np
import pandas as pd
from matplotlib import colors
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from tuprolog.solve.prolog import prolog_solver
from tuprolog.theory import Theory, mutable_theory

from psyke.extraction.hypercubic import HyperCubeExtractor
from psyke.utils.logic import data_to_struct, get_in_rule, get_not_in_rule

import matplotlib
#matplotlib.use('TkAgg')


def plot_init(xlim, ylim, xlabel, ylabel, size=(4, 3), equal=False):
    plt.figure(figsize=size)
    if equal:
        plt.gca().set_aspect(1)
    plt.xlim(xlim)
    plt.ylim(ylim)
    plt.gca().set_xlabel(xlabel)
    plt.gca().set_ylabel(ylabel)
    plt.gca().set_rasterized(True)


def plot_point(x, y, color, marker, ec=None):
    plt.scatter(x, y, c=color, marker=marker, edgecolors=ec, linewidths=0.6)


def plot_classification_samples(dataframe, classes, colors, markers, labels, loc, name, show=True, ec=None):
    marks = [Line2D([0], [0], color=c, marker=m, lw="0") for c, m in zip(colors, markers)]

    for cl, c, m in zip(classes, colors, markers):
        df = dataframe[dataframe.target == cl]
        plot_point(df["petal length"], df["petal width"], c, m, ec=ec)

    plt.gca().legend(marks, labels, loc=loc)
    plt.savefig("plot/{}.pdf".format(name), dpi=500, bbox_inches='tight')
    if show:
        plt.show()


def plot_boundaries(extractor: HyperCubeExtractor, x: str, y: str, colors: dict[str, str],
                    a: float = .5, h: str = '////////', ls='-', e=.05, fc='none', ec=None, reverse=False):
    cubes = extractor._hypercubes.copy()
    if reverse:
        cubes.reverse()
    for cube in cubes:
        plt.gca().fill_between((cube[x][0] - e, cube[x][1] + e), cube[y][0] - e, cube[y][1] + e,
                               fc=colors[cube.output] if fc is None else fc,
                               ec=colors[cube.output] if ec is None else ec, alpha=a, hatch=h, linestyle=ls)


def plot_surfaces(extractor: HyperCubeExtractor, x: str, y: str, colors: dict[str, str], ec='r', e=.05):
    for cube in extractor._hypercubes:
        plt.gca().fill_between((cube[x][0] - e, cube[x][1] + e), cube[y][0] - e, cube[y][1] + e,
                               fc='none', ec=ec)


def plot_perimeters(extractor: HyperCubeExtractor, x: str, y: str, colors: dict[str, str], n: int = 5,
                    ec: str = 'r', m: str = '*', s: int = 60, z: float = 1e10, lw: float = 0.8):
    for cube in extractor._hypercubes:
        for corner in cube.perimeter_samples(n):
            plt.scatter(corner[x], corner[y], c=colors[cube.output], marker=m, edgecolor=ec, s=s, zorder=z, linewidth=lw)


def plot_centers(extractor: HyperCubeExtractor, x: str, y: str, colors: dict[str, str],
                 ec: str = 'r', m: str = '*', s: int = 60, z: float = 1e10, lw: float = 0.8):
    for cube in extractor._hypercubes:
        center = cube.center
        plt.scatter(center[x], center[y], c=colors[cube.output], marker=m, edgecolor=ec, s=s, zorder=z, linewidth=lw)


def plot_corners(extractor: HyperCubeExtractor, x: str, y: str, colors: dict[str, str],
                 ec: str = 'r', m: str = '*', s: int = 60, z: float = 1e10, lw: float = 0.8):
    for cube in extractor._hypercubes:
        for corner in cube.corners():
            plt.scatter(corner[x], corner[y], c=colors[cube.output], marker=m, edgecolor=ec, s=s, zorder=z, linewidth=lw)


def plot_barycenters(extractor: HyperCubeExtractor, x: str, y: str, colors: dict[str, str],
                 ec: str = 'r', m: str = '*', s: int = 60, z: float = 1e10, lw: float = 0.8):
    for cube in extractor._hypercubes:
        center = cube.barycenter
        plt.scatter(center[x], center[y], c=colors[cube.output], marker=m, edgecolor=ec, s=s, zorder=z, linewidth=lw)


def predict_from_theory(theory: Theory, data: pd.DataFrame) -> list[float or str]:
    solver = prolog_solver(static_kb=mutable_theory(theory).assertZ(get_in_rule()).assertZ(get_not_in_rule()))
    index = data.shape[1] - 1
    y_element = data.iloc[0, -1]
    cast: Callable = lambda x: (str(x) if isinstance(y_element, str) else x)
    substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in data.iterrows()]
    return [cast(query.solved_query.get_arg_at(index)) if query.is_yes else -1 for query in substitutions]


def plot_theory(theory: Theory, data: pd.DataFrame = None, output: str = 'plot.pdf', azimuth: float = 45,
                distance: float = 9, elevation: float = 5, show_theory: bool = True, features: Iterable[str] = None) -> None:
    # Check if the number of common variables in clauses is less or equal to three.
    # If not raise an exception.
    fresh_theory = mutable_theory(theory)
    clauses = fresh_theory.clauses
    variables = sorted(list(set(arg.args[0].name.split('_')[0] for clause in clauses if clause.body_size > 0 and clause.body.is_recursive for arg in clause.body.unfolded)), reverse=True)
    if len(variables) > 3:
        raise Exception("Theory contains too many different features in the body of clauses, maximum is 3.")
    # If data is None, then create synthetic data covering a good portion of the variables space.
    # Just skip for now.
    if data is None:
        raise Exception("Method without data is not implemented yet")

    # Prepare data
    ys = predict_from_theory(fresh_theory, data)
    xs = data[variables].values.tolist()
    for i in range(len(ys)):
        xs[i].append(ys[i])

    # Prepare colors
    if isinstance(ys[0], str):
        np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)

        class ColorGenerator:

            def __init__(self):
                self.color_list = ['red', 'royalblue', 'green', 'orange', 'pink', 'acqua', 'grey']
                self.counter = 0

            def get_new_color(self) -> str:
                self.counter += 1
                if self.counter > len(self.color_list):
                    raise Exception("Classes exceed the maximum supported number (7)")
                return self.color_list[self.counter - 1]

        classes = set(ys)
        generator = ColorGenerator()
        class_color = {c: generator.get_new_color() for c in classes}
        get_color: Callable = lambda c: class_color[c]
    else:
        def color_fader(v: float = 0., c1: str = 'green', c2: str = 'red'):
            c1 = array(colors.to_rgb(c1))
            c2 = array(colors.to_rgb(c2))
            return colors.to_hex((1 - v) * c1 + v * c2)
        min_value = min(ys)
        max_value = max(ys)
        get_normalized_value: Callable = lambda v: (v - min_value)/(max_value - min_value)
        get_color: Callable = lambda c: color_fader(get_normalized_value(c))

    fig = plt.figure()
    fig.set_size_inches(10, 10)
    if len(variables) == 3:
        ax = fig.add_subplot(projection='3d')
    else:
        ax = fig.add_subplot()

    for x in xs:
        ax.scatter(*x[:-1], c=get_color(x[-1]), s=14)

    ax.set_xlabel(variables[0], fontsize=18)
    ax.set_ylabel(variables[1], fontsize=18)
    if len(variables) == 3:
        ax.set_zlabel(variables[2], fontsize=18)

    ax.azim = azimuth
    ax.dist = distance
    ax.elev = elevation
    ax.set_title('Predictions according to Prolog theory', fontsize=24)
    if show_theory:
        pass
        # ax.text2D(0., 0.88, pretty_theory(theory, new_line=False), transform=ax.transAxes, fontsize=8)
    if isinstance(ys[0], str):
        custom_lines = [Line2D([0], [0], marker='o', markerfacecolor=get_color(c),
                               markersize=20, color='w') for c in classes]
        ax.legend(custom_lines, classes, loc='upper left', numpoints=1, ncol=3, fontsize=18, bbox_to_anchor=(0, 0))
    plt.savefig(output, format='pdf')



from __future__ import annotations

import math
from abc import ABC
import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from sklearn.feature_selection import SelectKBest, f_regression, f_classif
from sklearn.linear_model import LinearRegression
from tuprolog.core import Var, Struct, clause
from tuprolog.theory import Theory, mutable_theory
from psyke.extraction import PedagogicalExtractor
from psyke.extraction.hypercubic.hypercube import HyperCube, RegressionCube, ClassificationCube, ClosedCube, Point, \
    GenericCube
from psyke.hypercubepredictor import HyperCubePredictor
from psyke.schema import Between, Outside, Value
from psyke.utils.logic import create_variable_list, create_head, to_var, Simplifier
from psyke.utils import Target
from psyke.extraction.hypercubic.strategy import Strategy, FixedStrategy


class HyperCubeExtractor(HyperCubePredictor, PedagogicalExtractor, ABC):
    def __init__(self, predictor, output, discretization=None, normalization=None):
        HyperCubePredictor.__init__(self, output=output, normalization=normalization)
        PedagogicalExtractor.__init__(self, predictor, discretization=discretization, normalization=normalization)
        self._default_surrounding_cube = False

    def _default_cube(self) -> HyperCube | RegressionCube | ClassificationCube:
        if self._output == Target.CONSTANT:
            return HyperCube()
        if self._output == Target.REGRESSION:
            return RegressionCube()
        return ClassificationCube()

    def _sort_cubes(self):
        cubes = [(cube.diversity, i, cube) for i, cube in enumerate(self._hypercubes)]
        cubes.sort()
        self._hypercubes = [cube[2] for cube in cubes]

    def extract(self, dataframe: pd.DataFrame) -> Theory:
        theory = PedagogicalExtractor.extract(self, dataframe)
        self._surrounding = HyperCube.create_surrounding_cube(dataframe, output=self._output)
        self._surrounding.update(dataframe, self.predictor)
        return theory

    def pairwise_fairness(self, data: dict[str, float], neighbor: dict[str, float]):
        cube1 = self._find_cube(data)
        cube2 = self._find_cube(neighbor)
        different_prediction_reasons = []

        if cube1.output == cube2.output:
            print("Prediction", cube1.output, "is FAIR")
        else:
            print("Prediction", cube1.output, "may be UNFAIR")
            print("It could be", cube2.output, "if:")
            for d in data:
                a, b = cube2.dimensions[d]
                if data[d] < a:
                    print('    ', d, 'increases above', round(a, 1))
                    different_prediction_reasons.append(d)
                elif data[d] > b:
                    print('    ', d, 'decreases below', round(b, 1))
                    different_prediction_reasons.append(d)
        return different_prediction_reasons

    def predict_counter(self, data: dict[str, float], verbose=True, only_first=True):
        output = ""
        prediction = None
        cube = self._find_cube(data)
        if cube is None:
            output += "The extracted knowledge is not exhaustive; impossible to predict this instance"
        else:
            prediction = self._predict_from_cubes(data)
            output += f"The output is {prediction}\n"

        point = Point(list(data.keys()), list(data.values()))
        cubes = self._hypercubes if cube is None else [c for c in self._hypercubes if cube.output != c.output]
        cubes = sorted([(cube.surface_distance(point), cube.volume(), i, cube) for i, cube in enumerate(cubes)])

        counter_conditions = []

        for _, _, _, c in cubes:
            if not only_first or c.output not in [o for o, _ in counter_conditions]:
                counter_conditions.append((c.output, {c: [val for val in v if val is not None and not val.is_in(
                    self.unscale(data[c], c))] for c, v in self.__get_conditions(data, c).items()}))

        if verbose:
            for o, conditions in counter_conditions:
                output += f"The output may be {o} if\n" + HyperCubeExtractor.__conditions_to_string(conditions)
            print(output)

        return prediction, counter_conditions

    @staticmethod
    def __conditions_to_string(conditions: dict[str, list[Value]]) -> str:
        output = ""
        for d in conditions:
            for i, condition in enumerate(conditions[d]):
                if i == 0:
                    output += f'     {d} is '
                else:
                    output += ' and '
                output += condition.print()
                if i + 1 == len(conditions[d]):
                    output += '\n'
        return output

    def __get_conditions(self, data: dict[str, float], cube: GenericCube) -> dict[str, list[Value]]:
        conditions = {d: [cube.interval_to_value(d, self.unscale)] for d in data.keys()
                      if d not in self._dimensions_to_ignore}
        for c in cube.subcubes(self._hypercubes):
            for d in conditions:
                condition = c.interval_to_value(d, self.unscale)
                if condition is None:
                    continue
                elif conditions[d][-1] is None:
                    conditions[d][-1] = -condition
                else:
                    try:
                        conditions[d][-1] *= -condition
                    except Exception:
                        conditions[d].append(-condition)
        return conditions

    def predict_why(self, data: dict[str, float], verbose=True):
        cube = self._find_cube(data)
        output = ""
        if cube is None:
            output += "The extracted knowledge is not exhaustive; impossible to predict this instance\n"
            if verbose:
                print(output)
            return None, {}
        prediction = self._predict_from_cubes(data)
        output += f"The output is {prediction} because\n"
        conditions = {c: [val for val in v if val is not None and val.is_in(self.unscale(data[c], c))]
                      for c, v in self.__get_conditions(data, cube).items()}

        if verbose:
            output += HyperCubeExtractor.__conditions_to_string(conditions)
            print(output)

        return prediction, conditions

    @staticmethod
    def _create_head(dataframe: pd.DataFrame, variables: list[Var], output: float | LinearRegression) -> Struct:
        return create_head(dataframe.columns[-1], variables[:-1], output) \
            if not isinstance(output, LinearRegression) else \
            create_head(dataframe.columns[-1], variables[:-1], variables[-1])

    def __drop(self, dataframe: pd.DataFrame):
        self._hypercubes = [cube for cube in self._hypercubes if cube.count(dataframe) > 1]

    def _create_theory(self, dataframe: pd.DataFrame) -> Theory:
        # self.__drop(dataframe)
        for cube in self._hypercubes:
            for dimension in cube.dimensions:
                if abs(cube[dimension][0] - self._surrounding[dimension][0]) < HyperCube.EPSILON * 2:
                    cube.set_infinite(dimension, '-')
                if abs(cube[dimension][1] - self._surrounding[dimension][1]) < HyperCube.EPSILON * 2:
                    cube.set_infinite(dimension, '+')

        if self._default_surrounding_cube:
            self._hypercubes[-1].set_default()

        new_theory = mutable_theory()
        for cube in self._hypercubes:
            variables = create_variable_list([], dataframe)
            variables[dataframe.columns[-1]] = to_var(dataframe.columns[-1])
            head = HyperCubeExtractor._create_head(dataframe, list(variables.values()),
                                                   self.unscale(cube.output, dataframe.columns[-1]))
            body = cube.body(variables, self._dimensions_to_ignore, self.unscale, self.normalization)
            new_theory.assertZ(clause(head, body))
        return HyperCubeExtractor._prettify_theory(new_theory)

    @staticmethod
    def _prettify_theory(theory: Theory) -> Theory:
        visitor = Simplifier()
        new_clauses = []
        for c in theory.clauses:
            body = c.body
            structs = body.unfolded if c.body_size > 1 else [body]
            new_structs = []
            for s in structs:
                new_structs.append(s.accept(visitor))
            new_clauses.append(clause(c.head, new_structs))
        return mutable_theory(new_clauses)


class FeatureRanker:
    def __init__(self, feat):
        self.scores = None
        self.feat = feat

    def fit(self, model, samples):
        predictions = np.array(model.predict(samples)).flatten()
        function = f_classif if isinstance(model, ClassifierMixin) else f_regression
        best = SelectKBest(score_func=function, k="all").fit(samples, predictions)
        self.scores = np.array(best.scores_) / max(best.scores_)
        return self

    def fit_on_data(self, samples):
        function = f_classif if isinstance(samples.iloc[0, -1], str) else f_regression
        best = SelectKBest(score_func=function, k="all").fit(samples.iloc[:, :-1], samples.iloc[:, -1])
        self.scores = np.array(best.scores_) / max(best.scores_)
        return self

    def rankings(self):
        return list(zip(self.feat, self.scores))


class Grid:
    def __init__(self, iterations: int = 1, strategy: Strategy | list[Strategy] = FixedStrategy()):
        self.iterations = iterations
        self.strategy = strategy

    def get(self, feature: str, depth: int) -> int:
        if isinstance(self.strategy, list):
            return self.strategy[depth].get(feature)
        else:
            return self.strategy.get(feature)

    def iterate(self) -> range:
        return range(self.iterations)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Grid ({}). {}".format(self.iterations, self.strategy)


class Node:
    def __init__(self, dataframe: pd.DataFrame, cube: ClosedCube = None):
        self.dataframe = dataframe
        self.cube: ClosedCube = cube
        self.right: Node | None = None
        self.left: Node | None = None

    @property
    def children(self) -> list[Node]:
        return [self.right, self.left]

    def search(self, point: dict[str, float]) -> ClosedCube:
        if self.right is None:
            return self.cube
        if point in self.right.cube:
            return self.right.search(point)
        return self.left.search(point)

    @property
    def leaves(self):
        if self.right is None:
            return 1
        return self.right.leaves + self.left.leaves



import itertools
import unittest
import pandas as pd
from sklearn.linear_model import LinearRegression

from psyke.extraction.hypercubic.hypercube import FeatureNotFoundException, ClosedRegressionCube, \
    ClosedClassificationCube, ClosedCube, ClassificationCube, RegressionCube, Point
from psyke.extraction.hypercubic.utils import MinUpdate, Expansion, ZippedDimension
from psyke.utils import get_int_precision
from sklearn.neighbors import KNeighborsRegressor
from psyke.extraction.hypercubic import HyperCube
from test import Predictor
from test.resources.datasets import get_dataset_path


class AbstractTestHypercube(unittest.TestCase):

    def setUp(self):
        self.x = (0.2, 0.6)
        self.y = (0.7, 0.9)
        self.dimensions = {'X': self.x, 'Y': self.y}
        self.mean = 0.5
        self.cube = HyperCube(self.dimensions, set(), self.mean)
        cubes = [({'X': (6.4, 7.9), 'Y': (5.7, 8.9)}, 5.3),
                 ({'X': (0.7, 0.8), 'Y': (0.75, 0.85)}, 6.1),
                 ({'X': (6.6, 7.0), 'Y': (9.1, 10.5)}, 7.5)]
        self.hypercubes = [HyperCube(cube[0], set(), cube[1]) for cube in cubes]
        self.dataset = pd.read_csv(get_dataset_path('arti'))
        self.filtered_dataset = self.dataset[self.dataset.apply(
            lambda row: (0.2 <= row['X'] < 0.6) & (0.7 <= row['Y'] < 0.9), axis=1)]

    def tuple_provider(self, closed=False):
        return (({'X': (self.x[0] + self.x[1]) / 2, 'Y': (self.y[0] + self.y[1]) / 2}, True),
                ({'X': self.x[0], 'Y': self.y[0]}, True),
                ({'X': self.x[0], 'Y': self.y[1]}, closed),
                ({'X': self.x[1], 'Y': self.y[0]}, closed),
                ({'X': self.x[1], 'Y': self.y[1]}, closed),
                ({'X': 1.5, 'Y': 3.6}, False))


class TestHypercube(AbstractTestHypercube):

    def test_dimension(self):
        self.assertEqual(self.dimensions, self.cube.dimensions)

    def test_limit_count(self):
        self.assertEqual(0, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(1, self.cube.limit_count)
        self.cube.add_limit('Y', '-')
        self.assertEqual(2, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(2, self.cube.limit_count)

    def test_output(self):
        self.assertEqual(self.mean, self.cube.output)

    def test_get(self):
        self.assertEqual(self.x, self.cube['X'])
        self.assertEqual(self.y, self.cube['Y'])
        with self.assertRaises(FeatureNotFoundException):
            _ = self.cube['Z']

    def test_get_first(self):
        self.assertEqual(self.x[0], self.cube.get_first('X'))
        self.assertEqual(self.y[0], self.cube.get_first('Y'))
        with self.assertRaises(FeatureNotFoundException):
            _ = self.cube.get_first('Z')

    def test_get_second(self):
        self.assertEqual(self.x[1], self.cube.get_second('X'))
        self.assertEqual(self.y[1], self.cube.get_second('Y'))
        with self.assertRaises(FeatureNotFoundException):
            _ = self.cube.get_second('Z')

    def test_copy(self):
        copy = self.cube.copy()
        self.assertEqual(self.cube.dimensions, copy.dimensions)
        self.assertEqual(self.cube.output, copy.output)
        self.assertIsInstance(copy, HyperCube)

    def test_expand(self):
        arguments = TestHypercube.expansion_provider()
        for arg in arguments:
            arg[0].expand(arg[1], self.hypercubes)
            self.assertEqual(arg[2], arg[0][arg[1].feature])

    def test_expand_all(self):
        updates = [MinUpdate('X', 0.2), MinUpdate('Y', 0.15)]
        surrounding = HyperCube({'X': (0.0, 0.8), 'Y': (0.1, 0.6)}, output=0.0)
        cube = HyperCube({'X': (0.1, 0.2), 'Y': (0.4, 0.4)}, output=0.4)
        cube.expand_all(updates, surrounding)
        self.assertEqual((0.0, 0.4), cube.dimensions['X'])
        self.assertEqual((0.25, 0.55), cube.dimensions['Y'])

    def test_overlap(self):
        self.assertIsNone(self.cube.overlap(self.hypercubes))
        self.assertFalse(self.cube.overlap(self.hypercubes[0]))
        self.assertFalse(self.cube.overlap(self.hypercubes[1]))
        self.cube.update_dimension('X', 0.6, 1.0)
        self.assertIsNotNone(self.cube.overlap(self.hypercubes))
        self.assertEqual(self.hypercubes[1], self.cube.overlap(self.hypercubes))
        self.assertFalse(self.cube.overlap(self.hypercubes[0]))
        self.assertTrue(self.cube.overlap(self.hypercubes[1]))
        self.assertTrue(self.hypercubes[1].overlap(self.cube))

    def test_has_volume(self):
        self.assertTrue(self.cube.has_volume())
        no_volume = self.cube.copy()
        no_volume.update_dimension('X', 1.0, 1.0)
        self.assertFalse(no_volume.has_volume())

    def test_equal(self):
        self.assertTrue(self.cube == self.cube)
        self.assertFalse(self.cube == self.hypercubes[1])

    def test_contains(self):
        arguments = self.tuple_provider()
        for arg in arguments:
            self.assertEqual(arg[1], arg[0] in self.cube)

    def test_count(self):
        self.assertEqual(self.dataset.shape[0], HyperCube.create_surrounding_cube(self.dataset).count(self.dataset))
        self.assertEqual(self.filtered_dataset.shape[0], self.cube.count(self.dataset))

    def test_create_samples(self):
        points = self.cube.create_samples(25)
        for k, v in self.cube.dimensions.items():
            self.assertTrue(all((points.loc[:, k] >= v[0]).values))
            self.assertTrue(all((points.loc[:, k] < v[1]).values))

    def test_add_limit(self):
        self.assertEqual(0, self.cube.limit_count)
        self.cube.add_limit('X', '-')
        self.assertEqual(1, self.cube.limit_count)
        self.cube.add_limit('X', '-')
        self.assertEqual(1, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(2, self.cube.limit_count)
        self.cube.add_limit('X', '+')
        self.assertEqual(2, self.cube.limit_count)

    def test_check_limits(self):
        self.assertIsNone(self.cube.check_limits('X'))
        self.cube.add_limit('X', '-')
        self.assertEqual('-', self.cube.check_limits('X'))
        self.cube.add_limit('X', '+')
        self.assertEqual('*', self.cube.check_limits('X'))
        self.assertIsNone(self.cube.check_limits('Y'))
        self.cube.add_limit('Y', '+')
        self.assertEqual('+', self.cube.check_limits('Y'))
        self.cube.add_limit('Y', '-')
        self.assertEqual('*', self.cube.check_limits('Y'))

    def test_filter_indices(self):
        expected = (self.dataset.X >= self.x[0]) & (self.dataset.X < self.x[1]) & \
                   (self.dataset.Y >= self.y[0]) & (self.dataset.Y < self.y[1])
        filtered = self.cube.filter_indices(self.dataset.iloc[:, :-1])
        self.assertTrue(all(expected == filtered))

    def test_filter_dataframe(self):
        expected = (self.dataset.X >= self.x[0]) & (self.dataset.X < self.x[1]) & \
                   (self.dataset.Y >= self.y[0]) & (self.dataset.Y < self.y[1])
        expected = self.dataset[expected].iloc[:, :-1]
        filtered = self.cube.filter_dataframe(self.dataset.iloc[:, :-1])
        self.assertTrue(all(expected == filtered))

    def test_update(self):
        model = KNeighborsRegressor()
        model.fit(self.dataset.iloc[:, :-1], self.dataset.iloc[:, -1])
        predictor = Predictor(model)
        self.cube.update(self.dataset, predictor)
        predictions = model.predict(self.dataset[
            (self.dataset.X >= self.x[0]) & (self.dataset.X < self.x[1]) &
            (self.dataset.Y >= self.y[0]) & (self.dataset.Y < self.y[1])
        ].iloc[:, :-1])
        self.assertEqual(self.cube.output, predictions.mean())
        self.assertEqual(self.cube.diversity, predictions.std())

    def test_update_dimension(self):
        new_lower, new_upper = 0.6, 1.4
        updated = {'X': (new_lower, new_upper),
                   'Y': (0.7, 0.9)}
        new_cube1 = self.cube.copy()
        new_cube1.update_dimension('X', new_lower, new_upper)
        self.assertEqual(updated, new_cube1.dimensions)
        new_cube2 = self.cube.copy()
        new_cube2.update_dimension('X', (new_lower, new_upper))
        self.assertEqual(updated, new_cube2.dimensions)

    def test_create_surrounding_cube(self):
        surrounding = HyperCube.create_surrounding_cube(self.dataset)
        for feature in self.dataset.columns[:-1]:
            self.assertEqual((round(min(self.dataset[feature]) - HyperCube.EPSILON * 2, get_int_precision()),
                              round(max(self.dataset[feature]) + HyperCube.EPSILON * 2, get_int_precision())),
                             surrounding.dimensions[feature])

    def test_cube_from_point(self):
        lower, upper, mean = 0.5, 0.8, 0.6
        cube = HyperCube.cube_from_point({'X': lower, 'Y': upper, 'z': mean})
        self.assertEqual({'X': (lower, lower), 'Y': (upper, upper)}, cube.dimensions)
        self.assertEqual(mean, cube.output)

    def test_check_overlap(self):
        self.assertTrue(HyperCube.check_overlap((self.hypercubes[0], ), (self.hypercubes[0].copy(), )))
        self.assertTrue(HyperCube.check_overlap(self.hypercubes, self.hypercubes + [self.hypercubes[0].copy()]))
        self.assertFalse(HyperCube.check_overlap(self.hypercubes, self.hypercubes))
        self.assertFalse(HyperCube.check_overlap(self.hypercubes[0:1], self.hypercubes[1:]))

    def test_init_diversity(self):
        d = 2.3
        self.cube.init_diversity(d)
        self.assertEqual(self.cube.diversity, d)

    def test_diversity(self):
        self.assertEqual(self.cube.diversity, 0.0)
        d = 56.3
        self.cube.init_diversity(d)
        self.assertEqual(self.cube.diversity, d)

    def test_volume(self):
        self.assertEqual(self.cube.volume(), (self.x[1] - self.x[0]) * (self.y[1] - self.y[0]))

    def test_diagonal(self):
        self.assertEqual(self.cube.diagonal(), ((self.x[1] - self.x[0])**2 + (self.y[1] - self.y[0])**2)**0.5)

    def test_center(self):
        self.assertEqual(self.cube.center, Point(list(self.dimensions.keys()),
                                                 [(val[0] + val[1]) / 2 for val in self.dimensions.values()]))

    def test_corners(self):
        self.assertEqual(self.cube.corners(), [
            Point(list(self.dimensions.keys()), values) for values in itertools.product(*self.dimensions.values())])

    def test_is_adjacent(self):
        cube_adj = HyperCube({'X': (0.6, 0.9), 'Y': self.y}, set(), self.mean)
        self.assertTrue(self.cube.is_adjacent(cube_adj))
        cube_not_adj = HyperCube({'X': self.y, 'Y': self.x}, set(), self.mean)
        self.assertFalse(self.cube.is_adjacent(cube_not_adj))

    def test_merge_along_dimension(self):
        cube_adj = HyperCube({'X': (0.6, 0.9), 'Y': self.y}, set(), self.mean)
        merged = self.cube.merge_along_dimension(cube_adj, 'X')
        self.assertEqual(merged['X'][0], 0.2)
        self.assertEqual(merged['X'][1], 0.9)

    def test_zip_dimensions(self):
        cube = HyperCube({'X': self.y, 'Y': self.x})
        expected = [ZippedDimension(d, self.cube[d], cube[d]) for d in self.dimensions]
        self.assertEqual(self.cube._zip_dimensions(cube), expected)

    def test_fit_dimension(self):
        new_dimensions = {'X': (5.2, 3.6), 'Y': (9.3, 6.4)}
        self.assertEqual(self.cube._fit_dimension(new_dimensions), new_dimensions)
        new_dimensions = {'Z': (5.2, 6.4)}
        self.assertEqual(self.cube._fit_dimension(new_dimensions), new_dimensions)

    @staticmethod
    def expansion_provider():
        cube1 = HyperCube({'X': (2.3, 6.4), 'Y': (8.9, 12.3)}, output=2.3)
        fake1 = cube1.copy()
        fake1.update_dimension('X', 0.5, 2.3)
        fake2 = cube1.copy()
        fake2.update_dimension('X', 6.4, 12.9)
        cube2 = cube1.copy()
        cube2.update_dimension('X', 9.5, 12.3)
        fake3 = cube2.copy()
        fake3.update_dimension('X', 5.0, 9.5)
        fake4 = cube2.copy()
        fake4.update_dimension('X', 12.3, 15.2)

        return [(cube1.copy(), Expansion(fake1, 'X', '-', 0.0), (0.5, 6.4)),
                (cube1.copy(), Expansion(fake2, 'X', '+', 0.0), (2.3, 6.6)),
                (cube2.copy(), Expansion(fake3, 'X', '-', 0.0), (7.0, 12.3)),
                (cube2.copy(), Expansion(fake4, 'X', '+', 0.0), (9.5, 15.2))]


class TestRegressionCube(AbstractTestHypercube):

    def test_copy(self):
        cube = RegressionCube(self.dimensions)
        copy = cube.copy()
        self.assertEqual(cube.dimensions, copy.dimensions)
        self.assertIsInstance(copy.output, LinearRegression)
        self.assertIsInstance(copy, RegressionCube)


class TestClassificationCube(AbstractTestHypercube):

    def test_copy(self):
        cube = ClassificationCube(self.dimensions)
        copy = cube.copy()
        self.assertEqual(cube.dimensions, copy.dimensions)
        self.assertEqual(cube.output, copy.output)
        self.assertIsInstance(copy, ClassificationCube)


class TestClosedCube(AbstractTestHypercube):

    def test_copy(self):
        cube = ClosedCube(self.dimensions)
        copy = cube.copy()
        self.assertEqual(cube.dimensions, copy.dimensions)
        self.assertEqual(cube.output, copy.output)
        self.assertIsInstance(copy, ClosedCube)

    def test_contains(self):
        cube = ClosedCube(self.dimensions)
        arguments = self.tuple_provider(True)
        for arg in arguments:
            self.assertEqual(arg[1], arg[0] in cube)


class TestClosedRegressionCube(AbstractTestHypercube):

    def test_copy(self):
        cube = ClosedRegressionCube(self.dimensions)
        copy = cube.copy()
        self.assertEqual(cube.dimensions, copy.dimensions)
        self.assertIsInstance(copy.output, LinearRegression)
        self.assertIsInstance(copy, ClosedRegressionCube)


class TestClosedClassificationCube(AbstractTestHypercube):

    def test_copy(self):
        cube = ClosedClassificationCube(self.dimensions)
        copy = cube.copy()
        self.assertEqual(cube.dimensions, copy.dimensions)
        self.assertEqual(cube.output, copy.output)
        self.assertIsInstance(copy, ClosedClassificationCube)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd
from enum import Enum

from sklearn.metrics import accuracy_score

from psyke import Extractor, Target
from psyke.extraction.hypercubic import Grid, FeatureRanker
from psyke.extraction.hypercubic.strategy import AdaptiveStrategy, FixedStrategy
from psyke.tuning import Objective, IterativeOptimizer, SKEOptimizer


class PEDRO(SKEOptimizer, IterativeOptimizer):
    class Algorithm(Enum):
        GRIDEX = 1,
        GRIDREX = 2,
        HEX = 3

    def __init__(self, predictor, dataframe: pd.DataFrame, max_error_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, max_depth: int = 3,
                 patience: int = 3, algorithm: Algorithm = Algorithm.GRIDREX, objective: Objective = Objective.MODEL,
                 output: Target = Target.CONSTANT, normalization=None, discretization=None):
        SKEOptimizer.__init__(self, predictor, dataframe, max_error_increase, min_rule_decrease,
                              readability_tradeoff, patience, objective, output, normalization, discretization)
        IterativeOptimizer.__init__(self, dataframe, max_error_increase, min_rule_decrease, readability_tradeoff,
                                    max_depth, patience, output, normalization, discretization)
        self.algorithm = Extractor.gridrex if algorithm == PEDRO.Algorithm.GRIDREX else \
            Extractor.gridex if algorithm == PEDRO.Algorithm.GRIDEX else Extractor.hex
        self.algorithm_name = "GridREx" if algorithm == PEDRO.Algorithm.GRIDREX else \
            "GridEx" if algorithm == PEDRO.Algorithm.GRIDEX else "HEx"
        self.ranked = FeatureRanker(dataframe.columns[:-1]).fit(predictor, dataframe.iloc[:, :-1]).rankings()
        predictions = self.predictor.predict(dataframe.iloc[:, :-1]).flatten()
        expected = self.dataframe.iloc[:, -1].values
        self.error = 1 - accuracy_score(predictions, expected) if output == Target.CLASSIFICATION else \
            abs(predictions - expected).mean()

    def _search_depth(self, strategy, critical, max_partitions):
        params, best = [], None

        for iterations in range(self.max_depth):
            current_params = self.__search_threshold(Grid(iterations + 1, strategy), critical, max_partitions)
            current_best = self._best(current_params)[1]
            print()
            best, to_break = self._check_iteration_improvement(best, current_best)
            params += current_params

            if len(params) > 1 and to_break:
                break
        return params

    def __search_threshold(self, grid, critical, max_partitions):
        step = self.error / 2.0
        threshold = self.error * 0.5
        params = []
        patience = self.patience
        while patience > 0:
            print("{}. {}. Threshold = {:.2f}. ".format(self.algorithm_name, grid, threshold), end="")
            extractor = self.algorithm(self.predictor, grid, min_examples=25, output=self.output,
                                       threshold=threshold, normalization=self.normalization)
            _ = extractor.extract(self.dataframe)
            error_function = (lambda *x: 1 - extractor.accuracy(*x)) if self.output == Target.CLASSIFICATION \
                else extractor.mae
            error, n = (error_function(self.dataframe, self.predictor) if self.objective == Objective.MODEL else
                        error_function(self.dataframe)), extractor.n_rules
            print("MAE = {:.2f}, {} rules".format(error, n))

            if len(params) == 0:
                params.append((error, n, threshold, grid))
                threshold += step
                continue

            if n > max_partitions:
                break

            if n == 1:
                params.append((error, n, threshold, grid))
                break

            if error > params[0][0] * self.max_error_increase:
                break

            improvement = (params[-1][0] / error) + (1 - n / params[-1][1])

            if improvement <= 1 or n > np.ceil(params[-1][1] * self.min_rule_decrease):
                patience -= 1
                step = max(step, abs(error - threshold) / max(patience, 1))
            elif not critical:
                patience = self.patience
            if error != params[-1][0] or n != params[-1][1]:
                params.append((error, n, threshold, grid))
            threshold += step
        return params

    def __contains(self, strategies, strategy):
        for s in strategies:
            if strategy.equals(s, self.dataframe.columns[:-1]):
                return True
        return False

    def search(self):
        max_partitions = 200
        base_partitions = FixedStrategy(2).partition_number(self.dataframe.columns[:-1]) * 3
        if base_partitions <= max_partitions:
            strategies = [FixedStrategy(2)]
            if FixedStrategy(3).partition_number(self.dataframe.columns[:-1]) <= max_partitions:
                strategies.append(FixedStrategy(3))
        else:
            strategies = []
            base_partitions = max_partitions

        for n in [2, 3, 5, 10]:
            for th in [0.99, 0.75, 0.67, 0.5, 0.3]:
                strategy = AdaptiveStrategy(self.ranked, [(th, n)])
                if strategy.partition_number(self.dataframe.columns[:-1]) < base_partitions and \
                        not self.__contains(strategies, strategy):
                    strategies.append(strategy)

        for (a, b) in [(0.33, 0.67), (0.25, 0.75), (0.1, 0.9)]:
            strategy = AdaptiveStrategy(self.ranked, [(a, 2), (b, 3)])
            if strategy.partition_number(self.dataframe.columns[:-1]) < base_partitions and \
                    not self.__contains(strategies, strategy):
                strategies.append(strategy)

        avg = 0.
        for strategy in strategies:
            avg += strategy.partition_number(self.dataframe.columns[:-1])
        avg /= len(strategies)

        params = []
        for strategy in strategies:
            params += self._search_depth(strategy,
                                         strategy.partition_number(self.dataframe.columns[:-1]) > avg,
                                         base_partitions)
        self.params = params

    def _print_params(self, name, params):
        print("**********************")
        print(f"Best {name}")
        print("**********************")
        print(f"Error = {params[0]:.2f}, {params[1]} rules")
        print(f"Threshold = {params[2]:.2f}")
        print(f"Iterations = {params[3].iterations}")
        print(f"Strategy = {params[3].strategy}")



from __future__ import annotations

from typing import Iterable

import numpy as np
import pandas as pd
from sklearn.neighbors import BallTree

from psyke import EvaluableModel, Target, get_int_precision
from psyke.extraction.hypercubic import RegressionCube, GenericCube, Point


class HyperCubePredictor(EvaluableModel):
    def __init__(self, output=Target.CONSTANT, discretization=None, normalization=None):
        super().__init__(discretization, normalization)
        self._hypercubes = []
        self._dimensions_to_ignore = set()
        self._output = output
        self._surrounding = None

    def _predict(self, dataframe: pd.DataFrame) -> Iterable:
        return np.array([self._predict_from_cubes(row.to_dict()) for _, row in dataframe.iterrows()])

    def _brute_predict(self, dataframe: pd.DataFrame, criterion: str = 'corner', n: int = 2) -> Iterable:
        predictions = np.array(self._predict(dataframe))
        idx = [prediction is None for prediction in predictions]
        if sum(idx) > 0:
            if criterion == 'default':
                predictions[idx] = np.array([HyperCubePredictor._get_cube_output(
                    self._surrounding, row
                ) for _, row in dataframe[idx].iterrows()])
            elif criterion == 'surface':
                predictions[idx] = np.array([HyperCubePredictor._get_cube_output(self._brute_predict_surface(row), row)
                                             for _, row in dataframe[idx].iterrows()])
            else:
                tree, cubes = self._create_brute_tree(criterion, n)
                predictions[idx] = np.array([HyperCubePredictor._brute_predict_from_cubes(
                    row.to_dict(), tree, cubes
                ) for _, row in dataframe[idx].iterrows()])
        return np.array(predictions)

    @staticmethod
    def _brute_predict_from_cubes(row: dict[str, float], tree: BallTree,
                                  cubes: list[GenericCube]) -> float | str:
        idx = tree.query([list(row.values())], k=1)[1][0][0]
        return HyperCubePredictor._get_cube_output(cubes[idx], row)

    def _brute_predict_surface(self, row: dict[str, float]) -> GenericCube:
        return min([(
            cube.surface_distance(Point(list(row.keys()), list(row.values()))), cube.volume(), cube
        ) for cube in self._hypercubes])[-1]

    def _create_brute_tree(self, criterion: str = 'center', n: int = 2) -> (BallTree, list[GenericCube]):
        admissible_criteria = ['surface', 'center', 'corner', 'perimeter', 'density', 'default']
        if criterion not in admissible_criteria:
            raise NotImplementedError(
                "'criterion' should be chosen in " + str(admissible_criteria)
            )

        points = [(cube.center, cube) for cube in self._hypercubes] if criterion == 'center' else \
            [(cube.barycenter, cube) for cube in self._hypercubes] if criterion == 'density' else \
            [(corner, cube) for cube in self._hypercubes for corner in cube.corners()] if criterion == 'corner' else \
            [(point, cube) for cube in self._hypercubes for point in cube.perimeter_samples(n)] \
            if criterion == 'perimeter' else None

        return BallTree(pd.concat([point[0].to_dataframe() for point in points], ignore_index=True)), \
            [point[1] for point in points]

    def _predict_from_cubes(self, data: dict[str, float]) -> float | str | None:
        cube = self._find_cube(data)
        if cube is None:
            return None
        elif self._output == Target.CLASSIFICATION:
            return HyperCubePredictor._get_cube_output(cube, data)
        else:
            return round(HyperCubePredictor._get_cube_output(cube, data), get_int_precision())

    def _find_cube(self, data: dict[str, float]) -> GenericCube | None:
        data = data.copy()
        for dimension in self._dimensions_to_ignore:
            if dimension in data:
                del data[dimension]
        for cube in self._hypercubes:
            if data in cube:
                return cube.copy()
        if self._hypercubes[-1].is_default:
            return self._hypercubes[-1].copy()

    @property
    def n_rules(self):
        return len(list(self._hypercubes))

    @property
    def volume(self):
        return sum([cube.volume() for cube in self._hypercubes])

    @staticmethod
    def _get_cube_output(cube, data: dict[str, float]) -> float:
        return cube.output.predict(pd.DataFrame([data])).flatten()[0] if \
            isinstance(cube, RegressionCube) else cube.output



from __future__ import annotations
from sklearn.model_selection import train_test_split
from tuprolog.solve.prolog import prolog_solver
from psyke.extraction.hypercubic import Grid, FeatureRanker
from psyke.utils.dataframe import get_discrete_dataset
from psyke.utils.logic import data_to_struct, get_in_rule, get_not_in_rule
from psyke.extraction.hypercubic.strategy import AdaptiveStrategy, FixedStrategy
from test import get_dataset, get_extractor, get_schema, get_model
from test.resources.tests import test_cases
from tuprolog.theory import Theory, mutable_theory
from tuprolog.theory.parsing import parse_theory
from typing import Callable
import ast
import numpy as np
from psyke import get_default_random_seed


def initialize(file: str) -> list[dict[str:Theory]]:
    for row in test_cases(file):
        params = dict() if row['extractor_params'] == '' else ast.literal_eval(row['extractor_params'])
        dataset = get_dataset(row['dataset'])

        # Dataset's columns are sorted due to alphabetically sorted extracted rules.
        # columns = sorted(dataset.columns[:-1]) + [dataset.columns[-1]]
        # dataset = dataset.reindex(columns, axis=1)

        training_set, test_set = train_test_split(dataset, test_size=0.05 if row['dataset'].lower() == 'house' else 0.5,
                                                  random_state=get_default_random_seed())

        schema, test_set_for_predictor = None, test_set
        if 'disc' in row.keys() and bool(row['disc']):
            schema = get_schema(training_set)
            params['discretization'] = schema
            training_set = get_discrete_dataset(training_set.iloc[:, :-1], schema) \
                .join(training_set.iloc[:, -1].reset_index(drop=True))
            test_set_for_predictor = get_discrete_dataset(test_set.iloc[:, :-1], schema) \
                .join(test_set.iloc[:, -1].reset_index(drop=True))

        # Handle Cart tests.
        # Cart needs to inspect the tree of the predictor.
        # Unfortunately onnx does not provide a method to do that.
        #if row['predictor'].lower() not in ['dtc', 'dtr']:
        #    params['predictor'] = Predictor.load_from_onnx(str(get_predictor_path(row['predictor'])))
        #else:
        predictor, fitted = get_model(row['predictor'], {})
        if not fitted:
            predictor.fit(training_set.iloc[:, :-1], training_set.iloc[:, -1])
        params['predictor'] = predictor

        # Handle GridEx tests
        # TODO: this is algorithm specific therefore it should be handled inside the algorithm itself.
        if 'grid' in row.keys() and bool:
            strategy, n = eval(row['strategies'])
            if strategy == "F":
                params['grid'] = Grid(int(row['grid']), FixedStrategy(n))
            else:
                ranked = FeatureRanker(training_set.columns[:-1]) \
                    .fit(params['predictor'], training_set.iloc[:, :-1]).rankings()
                params['grid'] = Grid(int(row['grid']), AdaptiveStrategy(ranked, n))

        extractor = get_extractor(row['extractor_type'], params)
        theory = extractor.extract(training_set)

        # Compute predictions from rules
        index = test_set.shape[1] - 1
        ordered_test_set = test_set.copy()
        ordered_test_set.iloc[:, :-1] = ordered_test_set.iloc[:, :-1].reindex(sorted(ordered_test_set.columns[:-1]),
                                                                              axis=1)
        cast, substitutions = get_substitutions(test_set, ordered_test_set, theory)
        expected = [cast(query.solved_query.get_arg_at(index)) for query in substitutions if query.is_yes]
        predictions = [prediction for prediction in extractor.predict(test_set_for_predictor.iloc[:, :-1])
                       if prediction is not None]

        yield {
            'extractor': extractor,
            'extracted_theory': theory,
            'extracted_test_y_from_theory': np.array(expected),
            'extracted_test_y_from_extractor': np.array(predictions),
            'test_set': test_set,
            'expected_theory': parse_theory(row['theory'] + '.') if row['theory'] != '' else None,
            'discretization': schema
        }


def get_substitutions(test_set, ordered_test_set, theory):
    cast: Callable = lambda x: (str(x) if isinstance(test_set.iloc[0, -1], str) else float(x.value))
    solver = prolog_solver(static_kb=mutable_theory(theory).assertZ(get_in_rule()).assertZ(get_not_in_rule()))
    substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in ordered_test_set.iterrows()]
    return cast, substitutions



from abc import ABC

import pandas as pd
from tuprolog.theory import Theory

from psyke import Extractor


class PedagogicalExtractor(Extractor, ABC):

    def __init__(self, predictor, discretization=None, normalization=None):
        Extractor.__init__(self, predictor=predictor, discretization=discretization, normalization=normalization)

    def extract(self, dataframe: pd.DataFrame) -> Theory:
        new_y = pd.DataFrame(self.predictor.predict(dataframe.iloc[:, :-1])).set_index(dataframe.index)
        data = dataframe.iloc[:, :-1].copy().join(new_y)
        data.columns = dataframe.columns
        return self._extract(data)

    def _extract(self, dataframe: pd.DataFrame) -> Theory:
        raise NotImplementedError('extract')



from abc import ABC
from typing import Iterable

from psyke import Clustering, Target
from psyke.extraction.hypercubic import HyperCube
from psyke.hypercubepredictor import HyperCubePredictor


class HyperCubeClustering(HyperCubePredictor, Clustering, ABC):

    def __init__(self, output: Target = Target.CONSTANT, discretization=None, normalization=None):
        HyperCubePredictor.__init__(self, output=output, discretization=discretization, normalization=normalization)

    def get_hypercubes(self) -> Iterable[HyperCube]:
        raise NotImplementedError('get_hypercubes')
